/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: partition-utils.h,v 1.9 2008/03/31 09:41:34 liuhui Exp $ */

#ifndef PHG_PARTITION_UTILS_H

typedef struct {
    FLOAT w;      /* weight              */
    FLOAT key;    /* key \in [0, 1]      */
    INT pn;       /* partition number    */
}DOTS;

/* load balance efficiency */
static const FLOAT SFC_LBE = 1.004;

/* max loops */
#define MAXLOOPS 16

#ifdef __cplusplus
extern "C" {
#endif

/* success 0, or fail, serial version now */
/* used by phgImport                      */
INT phgHilbertOrder(GRID *g);

/* 1d partitioner, partition [0, 1] to n parts*/
/* [0, p0), [p1, p2), ..., [p(n - 2), 1.]     */
/* such that weights of each part are equal   */
/* lenx, length of array x                    */
/* n, number of parts, should >= 1            */
/* return load balance efficiency fator       */
/*         ascending  order only              */
#if USE_MPI
FLOAT phgPartition1D(DOTS *x, INT lenx, INT n, MPI_Comm comm);
#endif

#if USE_MPI
/* remap partitions such that data migrations are minimized */
BOOLEAN phgPartitionRemap(GRID *g, INT nprocs, MPI_Comm comm);
#endif

/* print quality information */
void phgPartitionQuality(GRID *g);

#ifdef __cplusplus
}
#endif

#define PHG_PARTITION_UTILS_H
#endif	/* PHG_REMAP_H */
